##
## clirc -- An IRC client in Python.
##
## Copyleft 1998 Teemu Kalvas <chery@s2.org>
##
## User interface built on Gtkinter (0.3 on 0.99.8 at the moment but likely to
## change a lot).
##

import os
import string
import sys
import time
import traceback
from types import *

import Gtkinter
G = Gtkinter
del Gtkinter

import config
import reloader

## This module will be known as interface at runtime so we cannot actually
## import the modules which are going to be referring back to us, we trust
## the main driver to insert the appropriate module objects in this namespace
## by hand.
## import irc
## import policy

def do_quit(_x):
    G.mainquit()

def do_reload(_x):
    reloader._reload()

def build_about_dialog(_x):
    dialog = G.GtkDialog()
    dialog.set_title('About clirc')
    dialog.show()

    label = G.GtkLabel(' ')
    dialog.vbox.pack_start(label)
    label.show()

    hbox = G.GtkHBox(spacing=10)
    pixmap = G.GtkPixmap(dialog, config.libdir + '/clirc.xpm')
    label = G.GtkLabel(config.version)
    hbox.pack_start(pixmap)
    hbox.pack_start(label)
    dialog.vbox.pack_start(hbox)
    label.show()
    pixmap.show()
    hbox.show()

    hbox = G.GtkHBox(spacing=10)
    label = G.GtkLabel('Copyleft 1998 Teemu Kalvas <chery@s2.org>')
    hbox.pack_start(label, padding=10)
    dialog.vbox.pack_start(hbox, padding=10)
    label.show()
    hbox.show()

    button = G.GtkButton('OK')
    button.connect('clicked', lambda _x, dialog=dialog: dialog.destroy())
    dialog.action_area.pack_start(button, fill=0)
    button.show()
    dialog.show()

def build_exception_dialog():
    dialog = G.GtkDialog()
    dialog.set_title('Exception traceback')

    etype = sys.exc_type
    if not type(etype) == StringType:
	etype = etype.__name__
    hbox = G.GtkHBox()
    label = G.GtkLabel(etype + ': ' + str(sys.exc_value))
    hbox.pack_start(label, expand=0, padding=10)
    dialog.vbox.pack_start(hbox)
    label.show()
    hbox.show()
    for fname, linenum, fnname, src in traceback.extract_tb(sys.exc_traceback):
	hbox = G.GtkHBox()
	label = G.GtkLabel('File "' + fname + '", line ' + repr(linenum) +
			   ', in ' + fnname)
	hbox.pack_start(label, expand=0, padding=10)
	dialog.vbox.pack_start(hbox)
	label.show()
	hbox.show()
	hbox = G.GtkHBox()
	label = G.GtkLabel('    ' + src)
	hbox.pack_start(label, expand=0, padding=10)
	dialog.vbox.pack_start(hbox)
	label.show()
	hbox.show()

    button = G.GtkButton('Continue')
    button.connect('clicked', lambda _x, dialog=dialog: dialog.destroy())
    dialog.action_area.pack_start(button, fill=0)
    button.show()

    button = G.GtkButton('Terminate')
    button.connect('clicked', do_quit)
    dialog.action_area.pack_start(button, fill=0)
    button.show()
    dialog.show()

menus = [('File', [('Reload code', do_reload, ('r', 8)),
		   ('Quit', do_quit, ('q', 8))]),
	 ('Channel', [('Next', lambda _x: notebook.next_page(), ('.', 8)),
		      ('Previous', lambda _x: notebook.prev_page(),(',', 8))]),
	 ('Help', (), [('About', lambda _x: build_about_dialog(_x))])]

def _build_menus(shell, accelerator, template):
    if type(template) == ListType:
	for item in template:
	    _build_menus(shell, accelerator, item)
    elif type(template) == TupleType:
	menuitem = G.GtkMenuItem(template[0])
	for item in template[1:]:
	    if item == ():
		menuitem.right_justify()
	    elif type(item) == ListType:
		menu = G.GtkMenu()
		_build_menus(menu, accelerator, item)
		menuitem.set_submenu(menu)
	    elif callable(item):
		menuitem.connect('activate', item)
	    elif type(item) == TupleType and len(item) == 2:
		G._gtk.gtk_widget_install_accelerator(menuitem._o,
				 accelerator._at, 'activate', item[0], item[1])
	    else:
		#sys.stderr.write('Malformed menu item: ' + repr(item) + '\n')
		pass
	shell.append(menuitem)
	menuitem.show()
    else:
	#sys.stderr.write('Malformed menu template: ' + repr(template) + '\n')
	pass

def build_menus(menubar, accelerator):
    _build_menus(menubar, accelerator, menus)

def input_callback(sock, fn):
    return G.input_add(sock.fileno(), G.GDK.INPUT_READ, fn)

def output_callback(sock, fn):
    return G.input_add(sock.fileno(), G.GDK.INPUT_WRITE, fn)

def remove_callback(tag):
    G.input_remove(tag)

def current_key():
    point = notebook.current_page()
    for text, num, key, label in view_dict.values():
	if num == point:
	    return key
    #sys.stderr.write 'rrr in gtk_interface:current_key.'

class View:
    def __init__(sf, num):
	sf.num = num
	reloader.add_class(sf.__class__, 'interface', 'View')

    def check_crop(sf):
	text = notebook.children()[view_dict[sf.num][1]]
	if text.get_length() > 983040:
	    text.freeze()
	    i = 65536
	    while 1:
		str = text.get_chars(i, i + 4096)
		f = string.find(str, '\n')
		if f <> -1:
		    text.set_point(0)
		    text.forward_delete(i + f)
		    break
		i = i + 4096
	    text.set_point(text.get_length())
	    text.thaw()

    def insert_text(sf, str):
	sf.check_crop()
	text = notebook.children()[view_dict[sf.num][1]]
	text.insert_defaults(str + '\n')

    def replace_text(sf, fn, arg):
	sf.check_crop()
	text = notebook.children()[view_dict[sf.num][1]]
	i = text.get_length() - 1
	j = i
	while j > 0:
	    k = max(0, j - 4096)
	    str = text.get_chars(k, j)
	    f = string.rfind(str, '\n')
	    if f <> -1:
		str = text.get_chars(k + f + 1, i)
		replace, str = fn(str, arg)
		if replace:
		    text.freeze()
		    text.backward_delete(i - (k + f))
		    text.insert_defaults(str + '\n')
		    text.thaw()
		else:
		    text.insert_defaults(str + '\n')
		return
	    j = k
	text.insert_defaults(fn('', arg)[1] + '\n')

    def remove(sf):
	global notebook_pages

	view_dict[sf.num][0].destroy()
	notebook.remove_page(sf.num)
	notebook_pages = notebook_pages - 1
	limit = view_dict[sf.num][1]
	del view_dict[sf.num]
	for key, (text, num, _key, label) in view_dict.items():
	    if num > limit:
		view_dict[key] = (text, num - 1, _key, label)

def new_view(name, key):
    global view_num, notebook_pages

    text = G.GtkText()
    text.set_editable(0)

    ebox = G.GtkEventBox()
    label = G.GtkLabel(name)
    label.show()
    ebox.add(label)
    ebox.show()
    notebook.append_page(text, ebox)
    notebook.set_page(notebook_pages)
    notebook_pages = notebook_pages + 1
    text.show()
    text.realize()

    view_num = view_num + 1
    view_dict[view_num] = (text, notebook.current_page(), key, ebox)

    ebox.connect('button_press_event', lambda _x, _y, num=view_num:
		 notebook.set_page(view_dict[num][1]))

    return View(view_num)

def has_view_key(key):
    for t in view_dict.values():
	if t[2] == key:
	    return 1
    return 0

def entry_activate(entry):
    str = entry.get_text()
    entry.set_text('')
    try:
	policy.user_input(str)
    except:
	policy.exception()

def note_server_or_nick(net, *rest):
    label = view_dict[net.current_view()[0].num][3].children()[0]
    label.set(net.current_nick() + '@' + net.current_name())

def _reload():
    policy.exception.post(build_exception_dialog)
    policy.note_server_name.post(note_server_or_nick)
    policy.note_own_nick.post(note_server_or_nick)
    policy.note_topic.post(lambda net, id, chan, str:
			   tooltips.set_tip(view_dict[net.current_view()[1] \
					   [irc.lower(chan)].num][3], str, ''))
    policy.note_topic_.post(lambda net, chan, str:
			    tooltips.set_tip(view_dict[net.current_view()[1] \
					   [irc.lower(chan)].num][3], str, ''))

def timeout_handler():
    global timeout, timeout_tag

    try:
	G.timeout_remove(timeout_tag)
	timeout = timeout_fn()
	timeout_tag = G.timeout_add(int(1000.0 * (timeout - time.time())),
				    timeout_handler)
    except:
	policy.exception()

def set_timeout_handler(fn):
    global timeout_fn
    timeout_fn = fn

def timeout_latest(at):
    global timeout, timeout_tag

    if at - time.time() < 0.1:
	at = timeout_fn()
    if at < timeout:
	timeout = at
	G.timeout_remove(timeout_tag)
	timeout_tag = G.timeout_add(int(1000.0 * (at - time.time())),
				    timeout_handler)

def main(init):
    global view_num, notebook, notebook_pages, view_dict, timeout_fn, \
	   timeout, timeout_tag, accelerator, menubar, tooltips

    gtkrc = os.path.expanduser('~/.gtkrc')
    if os.path.exists(gtkrc):
	G.rc_parse(gtkrc)

    view_num = 0
    view_dict = {}

    win = G.GtkWindow()
    win.set_usize(640, 440)
    win.set_uposition(16, 64)
    win.connect('destroy', G.mainquit)
    win.connect('delete_event', G.mainquit)
    win.set_title(config.version)

    box = G.GtkVBox(0, 0)
    win.add(box)
    box.show()

    accelerator = G.GtkAcceleratorTable()
    win.add_accelerator_table(accelerator)
    menubar = G.GtkMenuBar()
    box.pack_start(menubar, expand=0)
    menubar.show()
    build_menus(menubar, accelerator)

    notebook = G.GtkNotebook()
    box.pack_start(notebook)
    notebook.set_tab_pos(G.POS_BOTTOM)
    notebook.set_show_tabs(1)
    notebook.set_show_border(1)
    notebook.show()
    notebook_pages = 0
    #notebook_prev = G.GtkButton()
    #notebook_prev.connect('clicked', lambda x: notebook.prev_page())
    #accelerator.install(notebook_prev, 'clicked', ord(','), 8)
    #notebook_next = G.GtkButton()
    #notebook_next.connect('clicked', lambda x: notebook.next_page())
    #accelerator.install(notebook_next, 'clicked', ord('.'), 8)

    entry = G.GtkEntry()
    box.pack_start(entry, expand=0)
    entry.connect('activate', entry_activate)
    entry.show()

    win.show()
    entry.grab_focus()

    tooltips = G.GtkTooltips()
    tooltips.enable()

    timeout = time.time() + 86400.0
    timeout_tag = G.timeout_add(86400000, timeout_handler)

    irc._start(init)
    policy._start()
    G.mainloop()

## End. ##
